import clang.cindex
from ctypes.util import find_library
clang.cindex.Config.set_library_file(find_library('clang'))

import os.path

#from stringify import *

link_name = None

def get_nodes(node, filename, kind_list):
    try:
        is_kind = (node.kind in kind_list)
    except ValueError as e:
        # unknown cursor kind
        is_kind = False

    if not is_kind:
        for c in node.get_children():
            for node in get_nodes(c, filename, kind_list):
                yield node
    else:
        if node.location.file is not None:
            path = os.path.dirname(filename)
            fname = os.path.dirname(str(node.location.file))
            is_in_dir = (os.path.commonprefix([fname, path]) == path)

            if is_in_dir:
                yield node

def get_function_arg_structs(all_functions):
    clang_types = clang.cindex.TypeKind

    is_struct = lambda arg: arg.get_canonical().get_pointee().get_canonical().kind is clang_types.RECORD
    get_struct_args = lambda fn: filter(is_struct, fn.type.argument_types())

    structs = map(get_struct_args, all_functions)
    structs = reduce(list.__add__, structs, [])

    structs = map(lambda n: n.get_pointee().get_declaration(), structs)

    return structs


def get_enums(node, filename):
    kind_list = [clang.cindex.CursorKind.ENUM_DECL]
    return get_nodes(node, filename, kind_list)

def get_macros(node, filename):
    kind_list = [clang.cindex.CursorKind.MACRO_DEFINITION]
    return get_nodes(node, filename, kind_list)


def get_functions(node, filename):
    kind_list = [clang.cindex.CursorKind.FUNCTION_DECL, clang.cindex.CursorKind.CXX_METHOD]
    return get_nodes(node, filename, kind_list)


def get_structs(node, filename):
    kind_list = [clang.cindex.CursorKind.STRUCT_DECL]
    
    nodes = get_nodes(node, filename, kind_list)
    structs = map(lambda n: n.type.get_declaration(), nodes)

    res = []
    for s in structs:
        if s not in res:
            res.append(s)
    return res

def get_typedefs(node, filename):
    kind_list = [clang.cindex.CursorKind.TYPEDEF_DECL]
    
    nodes = get_nodes(node, filename, kind_list)
    typedefs = map(lambda n: n.type.get_declaration(), nodes)

    res = []
    for t in typedefs:
        if t not in res:
            res.append(t)
    return res

def parent_path(node):
    parent = node.semantic_parent

    if parent is None:
        return []

    if parent.spelling is None:
        return []
    else:
        res = parent_path(parent) + [parent]
        return res

def canonical_function_name(node):
    res = ""
    for p in parent_path(node):
        if p.kind == clang.cindex.CursorKind.CLASS_DECL:
            res += "%s." % p.spelling
        elif p.kind == clang.cindex.CursorKind.NAMESPACE:
            res += "%s::" % p.spelling
        elif p.kind == clang.cindex.CursorKind.TRANSLATION_UNIT:
            pass
        else:
            assert(False)

    return res + node.spelling



import clang.cindex
import re

from parser import *

def rustify_variable_declaration(node):
    keywords = ['priv', 'loop', 'ref', 'in', 'type', 'where', 'impl', 'self', 'as', 'pub']

    name = node.spelling
    while name in keywords:
        name += "_"
    if len(name) == 0:
        name = "_"

    return "%s: %s" % (
            name,
#            rustify_type(node.type))
            rustify_type(node.type.get_canonical()))


def rustify_function_pointer(node):
    clang_types = clang.cindex.TypeKind

    canonical = node.get_canonical()
    canonical_pointee = canonical.get_pointee().get_canonical()

    assert canonical_pointee.kind is clang_types.FUNCTIONPROTO

    args = ", ".join(map(rustify_type, canonical_pointee.argument_types()))
    res = 'extern fn(%s)' % args

    if canonical_pointee.get_result().get_canonical().kind not in [clang_types.VOID, clang_types.INVALID]:
        res += " -> %s" % rustify_type(canonical_pointee.get_result().get_canonical())
    return "Option<%s>" % res # nullable


def rustify_pointer(node):
    clang_types = clang.cindex.TypeKind

    canonical = node.get_canonical()
    canonical_pointee = canonical.get_pointee().get_canonical()

    if canonical_pointee.kind == clang_types.FUNCTIONPROTO:
        return rustify_function_pointer(node)
    else:
        if canonical_pointee.is_const_qualified():
           return '*const ' + rustify_type(canonical_pointee)
        else:
           return '*mut ' + rustify_type(canonical_pointee)


def rustify_function_prototype(node):
    canonical = node.get_canonical()

    args = ",".join(map(rustify_type, canonical.argument_types()))
    ret = rustify_type(node.get_result().get_canonical())

    return "extern fn (%s) -> %s" % (args, ret)
#    return "libc::c_int/*FUNCTIONPROTO %s(%s)*/" % (ret,  args)
#        return "libc::c_int/*FUNCTIONPROTO %s*/" % rustify_function_declaration(node)


def rustify_type(node):
    clang_types = clang.cindex.TypeKind
    mapping = {
        clang_types.INT:       'libc::c_int',
        clang_types.UINT:      'libc::c_uint',
        clang_types.CHAR_S:    'libc::c_char',
        clang_types.CHAR_U:    'libc::c_uchar',
        clang_types.SHORT:     'libc::c_short',
        clang_types.USHORT:    'libc::c_ushort',
        clang_types.UCHAR:     'libc::c_uchar',
        clang_types.VOID:      'libc::c_void',
        clang_types.LONG:      'libc::c_long',
        clang_types.ULONG:     'libc::c_ulong',
        clang_types.ULONGLONG: 'libc::c_ulonglong',
        clang_types.LONGLONG:  'libc::c_longlong',
        clang_types.DOUBLE:    'libc::c_double',
        clang_types.FLOAT:     'libc::c_float',
        clang_types.ENUM:      'libc::c_uint',
    }

    canonical = node.get_canonical()
    canonical_kind = canonical.kind
    if canonical_kind == clang_types.POINTER:
        return rustify_pointer(node)
    elif canonical_kind is clang_types.RECORD:
        return re.sub('^struct ', '', re.sub('^const ', '', canonical.spelling))
    elif canonical_kind == clang_types.FUNCTIONPROTO:
        return rustify_function_prototype(node)
    elif canonical_kind == clang_types.INCOMPLETEARRAY:
        #return 'libc::c_int /* INCOMPLETEARRAY */'
        return "*mut %s /* INCOMPLETEARRAY */" % rustify_type(canonical.get_array_element_type())
    elif canonical_kind == clang_types.CONSTANTARRAY:
        return "[%s; %i]" % (rustify_type(canonical.get_array_element_type()),
                canonical.get_array_size())
    elif canonical_kind in mapping:
        return mapping[canonical_kind]
    else:
        #raise Exception("Not implemented: Type=%s" % canonical_kind)
        return "%s" %   node.get_canonical().kind


def rustify_function_declaration(node, link_name=None):
    args = ", ".join(map(rustify_variable_declaration, node.get_arguments()))

    res = "pub fn %s(%s)" % (canonical_function_name(node), args)

    if node.result_type.get_canonical().kind is not clang.cindex.TypeKind.VOID:
        res += ' -> %s' % rustify_type(node.result_type.get_canonical())

    res = 'extern "C" {\n\t%s;\n}\n' % res
    if link_name is not None:
        res = '#[link(name="%s")]\n%s' % (link_name, res)

    return res


def rustify_struct_declaration(node):
    canonical = node.type.get_canonical().get_declaration()
    node = node.type.get_declaration()
    res =  "#[repr(C)]\n"
    res += "pub struct %s" % node.spelling

    members = " {\n"
    for c in canonical.get_children():
        members += "\t%s,\n" % rustify_variable_declaration(c)

    if len(members) > 3:
        members += "}\n"
        res += members
    else:
        res += ';\n'
    return res

def rustify_enum_declaration(node):
    name = node.spelling
    if len(name) == 0:
        name = node.referenced.type.spelling

    if len(name) > 0:
        all_positive = all(map(lambda c: c.enum_value > -1, node.get_children()))

        if all_positive:
            typ = ("u32", "libc::c_uint")
        else:
            typ = ("i32", "libc::c_int")

        res = "bitflags! {\n" % name
        res += "\tflags %s: %s {\n" % (name, typ[1])
        for c in node.get_children():
            res += "\t\tconst %s =\t%i as %s,\n" % (c.spelling, c.enum_value, typ[1])
        res += "\t}\n"
        res += "\n"
    else:
        res = ""
        for c in node.get_children():
            res += "pub const %s: i32 = %i;\n" % (c.spelling, c.enum_value)

    return res

def rustify_macro_declaration(macro):
    val = ''.join(map(lambda x: x.spelling, list(macro.get_tokens())[1:-1]))

    for base in [10, 16]:
        try:
            return "pub const %s: i32 = %i;\n" % (macro.displayname, int(val,base))
        except ValueError:
            pass

    return ""



#!/usr/bin/env python
""" Usage: call with <filename> <typename>
"""

import sys
import sets
import itertools
import clang.cindex

from ctypes.util import find_library
clang.cindex.Config.set_library_file(find_library('clang'))

from rustify import *
from stringify import *

header = """
extern crate libc;

#[macro_use]
extern crate bitflags;

use std::mem;

"""

def unique(input):
    output = []
    for el in input:
        if el not in output:
            output.append(el)
    return output

def main(args=sys.argv):

    index = clang.cindex.Index.create()

    args = args[1:]
    link_name = args.pop(0)
    fname = args.pop(0)

    PARSE_SKIP_FUNCTION_BODIES=64
    PARSE_DETAILED_PROCESSING_RECORD=1
    PARSE_INCOMPLETE=2

    options = PARSE_SKIP_FUNCTION_BODIES | PARSE_INCOMPLETE | PARSE_DETAILED_PROCESSING_RECORD
    tu = index.parse(fname, args, options=options)

    print header

    all_functions1, all_functions2 = itertools.tee(get_functions(tu.cursor, tu.spelling))

    all_structs = unique(get_structs(tu.cursor, tu.spelling) +
                            get_function_arg_structs(all_functions1))
    for struct in all_structs:
        print "/*\n%s*/" % stringify_struct_declaration(struct)
        print rustify_struct_declaration(struct)

    for func in all_functions2:
        print "/*\n%s*/" % stringify_function_declaration(func)
        print rustify_function_declaration(func, link_name=link_name)
        print

    all_enums = get_enums(tu.cursor, tu.spelling)
    for enum in all_enums:
        print "/*\n%s*/" % stringify_enum_declaration(enum)
        print rustify_enum_declaration(enum)

    all_macros = get_macros(tu.cursor, tu.spelling)
    for m in all_macros:
        print "/* %s */" % stringify_macro_declaration(m)
        print rustify_macro_declaration(m)

"""
    get_arguments = lambda arg: arg._tu
    all_args = map(lambda fn: map(get_arguments, fn.get_arguments()), get_functions(tu.cursor, tu.spelling))
    #print set(reduce(lambda a,b: a+b, all_args))
"""

if __name__ == "__main__":
    main()



import os
import sys
import re

from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
NAME = 'cxx2rs'

with open(os.path.join(here, 'README.rst')) as readme:
    README = readme.read()

with open(os.path.join(here, 'CHANGES.rst')) as changes:
    CHANGES = changes.read()

with open(os.path.join(here, NAME, '__init__.py')) as version:
    VERSION = re.compile(r".*__version__ = '(.*?)'",
                         re.S).match(version.read()).group(1)



requires = [
    'setuptools',
    'clang'
    ]

setup(name=NAME,
      version=VERSION,
      description='A rust-binding generator for C/C++ files',
      long_description=README + '\n\n' + CHANGES,
      classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Intended Audience :: Developers",
        ],
      author='manuels',
      url='https://github.com/manuels/cxx2rs',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      entry_points="""\
      [console_scripts]
      {name} = {name}.__main__:main
      """.format(name=NAME),
      )

